unit Deriver;

{******************************************************************************
 Nome:			Deriver
 Scopo:     Polymorfic creation of objects.
 Generalit:Deriver defines the TDeriver class. Other class trees can use 
            TDeriver to obtain a polymorfic creation based on criteria
            evaluated at run time.
 Autore:    Ing. M. Venturini
 Revisore:
 Data:      24 Feb. 97
 Revisioni:
******************************************************************************}

interface

uses
	SysUtils, Classes;

type
	TDeriverMatchMethod = function (const Info: TObject): Boolean of object;
  TDeriverNoMatch = procedure(			Sender: TObject;
  														const Info: TObject) of object;
  TDeriver = class(TObject)
  private
  	Derived: TStringList;
    FOnNoMatch: TDeriverNoMatch;
  public
  	property OnNoMatch: TDeriverNoMatch read FOnNoMatch write FOnNoMatch;
  	procedure Register(const DerivedClass: TClass;
    									 const MatchMethod: TDeriverMatchMethod);
  	function DerivedClass(const Info: TObject): TClass;
    constructor Create;
    destructor Destroy; override;
  end;
  EDeriver = class(Exception);
  EMatch = class(EDeriver);

{*****************************************************************************}

implementation

const
	EMatchMsg = ' - No derived class matched';

var
	I: Integer;
	DeriverList: TStringList; {List of TDeriver instances}

{* TDerived ******************************************************************}
{Derived is a list of TDerived istances that contains the class reference and
 the address of the method to be used to determine the class corresponding to
 a set of creational infos                                                    }

type
  TDerived = class(TObject)
  private
    DerivedClass: TClass;
    MatchMethod: TDeriverMatchMethod;
  public
  	constructor Create(const DerivedClass: TClass;
                       const MatchMethod: TDeriverMatchMethod);
  end;

constructor TDerived.Create(const DerivedClass: TClass;
										        const MatchMethod: TDeriverMatchMethod);
begin

	inherited Create;
  Self.DerivedClass := DerivedClass;
  Self.MatchMethod := MatchMethod;

end;

{* TDeriver ******************************************************************}

{******************************************************************************
 --	procedure Register [method]
 Register a new derived class.
 --	Parameters:
 		DerivedClass:   in TClass
    								Class reference to be registered.
    MatchMethod:    in TMatchMethod
                    Method of DerivedClass to be used to check it against
                    condition data.
******************************************************************************}

procedure TDeriver.Register(const DerivedClass: TClass;
														const MatchMethod: TDeriverMatchMethod);
begin

  Derived.AddObject(DerivedClass.ClassName,TDerived.Create(DerivedClass,MatchMethod));

end;

{******************************************************************************
 --	function DerivedClass [method]
 For every registered class the MatchMethod is called and if it returns true
 its class reference is returned.
 --	Parameters:
 		Info:	      		in TObject
										Condition data, it is passed to MatchMethod.
 		Result:         out TClass
    								Class reference whose MatchMethod returned true.
 -- Exception:			EMatch
 										Raised if no registerd class match the condition data and
                    if OnNoMatch event is not assigned.
******************************************************************************}

function TDeriver.DerivedClass(const Info: TObject): TClass;

var
	I: Integer;

begin

	I := 0;
 	repeat
    with Derived.Objects[I] as TDerived do
	    if MatchMethod(Info) then begin
	    	Result :=	DerivedClass;
  	    Exit;
    	end;
    Inc(I);
 	until I >= Derived.Count;
  if Assigned(FOnNoMatch) then begin
  	FOnNoMatch(Self,Info);
    Result := nil;
  end
  else
	  raise EMatch.Create(Derived.Strings[0] + EMatchMsg);

end;

{******************************************************************************
 --	procedure Create [constructor]
 Class constructor. Create Derived list and register itself.
******************************************************************************}

constructor TDeriver.Create;
begin

	inherited Create;
	Derived := TStringList.Create;
  DeriverList.AddObject('',Self);

end;

{******************************************************************************
 --	procedure Create [destructor]
 Class destructor. Release Derived contents and Derived itself.
******************************************************************************}

destructor TDeriver.Destroy;

var
	I: Integer;

begin

	for I := Derived.Count - 1 downto 0 do
  	(Derived.Objects[I] as TDerived).Free;
  Derived.Free;
  inherited Destroy;

end;

initialization

	DeriverList := TStringList.Create;

finalization

	for I := DeriverList.Count - 1 downto 0 do
  	(DeriverList.Objects[I] as TDeriver).Free;
  DeriverList.Free;

end.
